import ast
import astor
import black
from typing import List, Dict

from mlopx import Component
from mlopx.consts import (
    IMPORTS_MAPPING,
    KFP_PIPELINE_DECORATOR,
    PIPELINE_IMPORTS,
)


class PipelineBuilder:

    def __init__(self):
        self.tree = ast.Module(body=[])
        self.func_node = None


    def add_imports(self, components: List[Component]):
        """
        Add imports to the pipeline file
        """
        # KFP imports
        imports = {}
        for name in PIPELINE_IMPORTS:
            module, names = IMPORTS_MAPPING[name]
            imports[module] = imports.get(module, []) + names

        for module, names in imports.items():
            node = ast.ImportFrom(
                module=module,
                names=[ast.alias(name=name, asname=None) for name in names],
                level=0,
            )
            self.tree.body.append(node)

        # Component imports
        for component in components:
            node = ast.ImportFrom(
                module=f"kfp_{component.name}",
                names=[ast.alias(name=component.name, asname=None)],
                level=0,
            )
            self.tree.body.append(node)

        return self


    def create_function(self, func_name: str):
        """
        Create the pipeline function
        """
        self.func_node = ast.FunctionDef(
            name=func_name,
            args=ast.arguments(
                args=[],
                defaults=[],
            ),
            body=[],
        )
        return self


    def add_decorator(self, pipeline_name: str):
        """
        Add the kfp pipeline decorator to the function
        """
        self.func_node.decorator_list = [
            ast.Call(
                func=ast.Name(id=KFP_PIPELINE_DECORATOR, ctx=ast.Load()),
                args=[],
                keywords=[ast.keyword(arg="name", value=ast.Constant(s=pipeline_name))],
            )
        ]
        return self


    def call_components(self, components: List[Component], artifacts: Dict):
        """
        Call the compiled components in the pipeline function
        """
        for component in components:
            args = {}
            for arg_name, arg_type in component.arg_types.items():
                if arg_name in component.user_args:
                    args[arg_name] = ast.Constant(component.user_args[arg_name])
                elif arg_type.startswith("Input"):
                    args[arg_name] = ast.Subscript(
                        value=ast.Attribute(
                            value=ast.Name(
                                id=f"{artifacts[arg_name]}_task", ctx=ast.Load()
                            ),
                            attr="outputs",
                            ctx=ast.Load(),
                        ),
                        slice=ast.Constant(value=arg_name),
                        ctx=ast.Load(),
                    )

            node = ast.Assign(
                targets=[ast.Name(id=f"{component.name}_task", ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Name(id=component.name, ctx=ast.Load()),
                    args=[],
                    keywords=[ast.keyword(arg=k, value=v) for k, v in args.items()],
                ),
            )
            self.func_node.body.append(node)

        self.tree.body.append(self.func_node)
        return self


    def mount_volumes(self, components: List[Component]):
        """
        Mount volumes to components
        """
        for component in components:
            for pvc, mount_path in component.volumes:
                invoke_node = ast.Call(
                    func=ast.Name(id="mount_pvc", ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(
                            arg="task",
                            value=ast.Name(id=f"{component.name}_task", ctx=ast.Load()),
                        ),
                        ast.keyword(
                            arg="pvc_name",
                            value=ast.Constant(value=pvc)
                        ),
                        ast.keyword(
                            arg="mount_path",
                            value=ast.Constant(value=mount_path)
                        )
                    ],
                )
                node = ast.Assign(
                    targets=[ast.Name(id=f"{component.name}_task", ctx=ast.Store())],
                    value=invoke_node,
                )
                self.func_node.body.append(node)
        return self

        
    def add_node_selector(self, components: List[Component], node_selectors: List[str]):
        """
        Add node selectors to components
        """
        if not node_selectors:
            return self
        
        for i, component in enumerate(components):
            invoke_node = ast.Call(
                func=ast.Name(id="add_node_selector", ctx=ast.Load()),
                args=[],
                keywords=[
                    ast.keyword(
                        arg="task",
                        value=ast.Name(id=f"{component.name}_task", ctx=ast.Load()),
                    ),
                    ast.keyword(
                        arg="label_key",
                        value=ast.Constant(value="kubernetes.io/hostname")
                    ),
                    ast.keyword(
                        arg="label_value",
                        value=ast.Constant(value=node_selectors[i])
                    )
                ],
            )
            node = ast.Assign(
                targets=[ast.Name(id=f"{component.name}_task", ctx=ast.Store())],
                value=invoke_node,
            )
            self.func_node.body.append(node)
        return self


    def create_client(self, kfp_url: str):
        """
        Create the kfp client
        """
        node = ast.Assign(
            targets=[ast.Name(id="client", ctx=ast.Store())],
            value=ast.Call(
                func=ast.Name(id="Client", ctx=ast.Load()),
                args=[],
                keywords=[ast.keyword(arg="host", value=ast.Constant(value=kfp_url))],
            ),
        )
        self.tree.body.append(node)
        return self


    def add_create_run(self, func_name: str, enable_caching: bool):
        """
        Call the kfp create run function to run the created pipeline
        """
        node = ast.Assign(
            targets=[ast.Name(id="run", ctx=ast.Store())],
            value=ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="client", ctx=ast.Load()),
                    attr="create_run_from_pipeline_func",
                    ctx=ast.Load(),
                ),
                args=[],
                keywords=[
                    ast.keyword(
                        arg="pipeline_func",
                        value=ast.Name(id=func_name, ctx=ast.Load()),
                    ),
                    ast.keyword(
                        arg="enable_caching", value=ast.Constant(value=enable_caching)
                    ),
                ],
            ),
        )
        print_node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="print", ctx=ast.Load()),
                args=[
                    ast.Constant(value="Run ID: "),
                    ast.Attribute(
                        value=ast.Name(id="run", ctx=ast.Load()),
                        attr="run_id",
                        ctx=ast.Load(),
                    ),
                ],
                keywords=[],
            )
        )
        self.tree.body.extend([node, print_node])
        return self


    def save_pipeline(self) -> None:
        """
        Save the pipeline to a file
        """
        with open("kfp_pipeline.py", "w") as f:
            ast.fix_missing_locations(self.tree)
            kfp_pipeline = astor.to_source(self.tree)
            kfp_pipeline = black.format_str(kfp_pipeline, mode=black.Mode())
            f.write(kfp_pipeline)



class OutputDataset:
    """Marker type for output datasets."""


class InputDataset:
    """Marker type for input datasets."""


class InputModel:
    """Marker type for input models."""


class OutputModel:
    """Marker type for output models."""



import ast
import astor
import black
from typing import Dict

from mlopx.consts import IMPORTS_MAPPING, TYPES_MAPPING, KFP_COMPONENT_DECORATOR


class ComponentConverter:

    def __init__(self, tree: ast.Module):
        self.tree = tree


    def remove_type_imports(self):
        """
        Remove marked types imports
        """
        import_nodes = []
        for node in ast.walk(self.tree):
            if isinstance(node, ast.ImportFrom) and "artifacts" in node.module:
                import_nodes.append(node)
        for node in import_nodes:
            self.tree.body.remove(node)
        return self


    def get_imports(self, arg_types: Dict) -> Dict:
        """
        Get modules and names to import
        """
        imports = {}
        for arg_type in arg_types.values():
            if arg_type not in IMPORTS_MAPPING:
                continue
            module, names = IMPORTS_MAPPING[arg_type]
            if module not in imports:
                imports[module] = set()
            imports[module].update(names)
        return imports


    def add_imports(self, arg_types):
        """
        Add import statements to the top of the file
        """
        imports = self.get_imports(arg_types)
        for module, names in imports.items():
            node = ast.ImportFrom(
                module=module,
                names=[ast.alias(name=name, asname=None) for name in names],
                level=0,
            )
            self.tree.body.insert(0, node)

        # Import @dsl.component decorator
        module, names = IMPORTS_MAPPING[KFP_COMPONENT_DECORATOR]
        node = ast.ImportFrom(
            module=module,
            names=[ast.alias(name=name, asname=None) for name in names],
            level=0,
        )
        self.tree.body.insert(0, node)
        return self


    def add_decorator(self, component_name: str, image: str):
        """
        Add the kfp component decorator to the function
        """
        for node in ast.walk(self.tree):
            if isinstance(node, ast.FunctionDef) and node.name == component_name:
                decorator_node = ast.Call(
                    func=ast.Name(id=KFP_COMPONENT_DECORATOR, ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(arg="base_image", value=ast.Constant(s=image))
                    ],
                )
                node.decorator_list.append(decorator_node)
                break
        return self


    def update_arg_types(self, component_name: str):
        """
        Replace artifact marker types with kfp data types
        """
        for node in ast.walk(self.tree):
            if isinstance(node, ast.FunctionDef) and node.name == component_name:
                for arg in node.args.args:
                    arg_type = arg.annotation.id
                    if arg_type in TYPES_MAPPING:
                        arg.annotation.id = TYPES_MAPPING[arg_type]
                break
        return self


    def save_component(self, filename: str):
        """
        Save the converted component to a file
        """
        with open(f"kfp_{filename}", "w") as f:
            ast.fix_missing_locations(self.tree)
            kfp_component = astor.to_source(self.tree)
            kfp_component = black.format_str(kfp_component, mode=black.Mode())
            f.write(kfp_component)



import os
import requests
import ast
import astor
import black
from pathlib import Path
from typing import List, Tuple

from mlopx import Component, PipelineBuilder
from mlopx.consts import ARGPARSE_CODE, METADATA_FILENAME


class Pipeline:

    def __init__(self, name: str):
        self.name = name
        self.func_name = name.replace(" ", "_").lower()
        self.components = []
        self.artifacts = {}
        self.tree = None


    def add(self, components: List[Component]) -> None:
        """
        Add a list of components to the pipeline
        """
        self.components.extend(components)
        for component in components:
            for arg_name, arg_type in component.arg_types.items():
                if "Output" in arg_type:
                    self.artifacts[arg_name] = component.name

    
    def create_tmp_pipeline(self, tmp_filename: str) -> None:
        """
        Create a temporary pipeline file to be submitted
        """
        with open("pipeline.py", "r") as f:
            code = f.read()
            tree = ast.parse(code)

        argparse_nodes = [n for line in ARGPARSE_CODE for n in ast.parse(line).body]
        run_node = ast.parse("pipeline.build(args.u, args.c, args.p)").body

        for i, node in enumerate(tree.body):
            if not isinstance(node, ast.ImportFrom) and not isinstance(node, ast.Import):
                break
            
        tree.body = tree.body[:i] + argparse_nodes + tree.body[i:-1] + run_node
        with open(tmp_filename, "w") as f:
            ast.fix_missing_locations(tree)
            kfp_pipeline = astor.to_source(tree)
            kfp_pipeline = black.format_str(kfp_pipeline, mode=black.Mode())
            f.write(kfp_pipeline)


    def prepare_files(self) -> List[Tuple]:
        """
        Prepare the files for submission
        """
        # Component files
        files = [
            ("components", (c.filename, open(c.filename, "rb"))) for c in self.components
        ]

        # Metadata file
        if METADATA_FILENAME not in os.listdir():
            raise FileNotFoundError("You must have a metadata.json file in the current directory")
        files.append(("metadata", (METADATA_FILENAME, open(METADATA_FILENAME, "rb"))))

        # Pipeline file
        tmp_file = "pipeline_tmp.py"
        self.create_tmp_pipeline(tmp_file)
        files.append(("pipeline", ("pipeline.py", open(tmp_file, "rb"))))
        os.remove(tmp_file)

        return files


    def send_pipeline(self, server_url: str, files: List[Tuple]) -> requests.Response:
        """
        Send the pipeline files to the server
        """
        try:
            response = requests.post(f"{server_url}/submit/", files=files)
            response.raise_for_status()
            return response
        except requests.RequestException as e:
            print(f"An error occurred: {e}")
            return None


    def handle_response(self, response: requests.Response) -> None:
        """
        Handle the server response
        """
        if response:
            try:
                print(response.json())
            except ValueError:
                print("Failed to parse response JSON")


    def submit(self, server_url: str) -> None:
        """
        Submit the pipeline to the server
        """
        files = self.prepare_files()
        response = self.send_pipeline(server_url, files)
        self.handle_response(response)


    def build(self, kfp_url: str, enable_caching: str, placement: List[str]):
        """
        Build the kfp pipeline
        """
        for component in self.components:
            component.convert()

        builder = PipelineBuilder()
        (
            builder.add_imports(self.components)
            .create_function(self.func_name)
            .add_decorator(self.name)
            .call_components(self.components, self.artifacts)
            .mount_volumes(self.components)
            .add_node_selector(self.components, placement)
            .create_client(kfp_url)
            .add_create_run(self.func_name, enable_caching)
            .save_pipeline()
        )



KFP_COMPONENT_DECORATOR = "dsl.component"
KFP_PIPELINE_DECORATOR = "dsl.pipeline"

PIPELINE_IMPORTS = ["dsl.pipeline", "mount_pvc", "add_node_selector", "client"]

IMPORTS_MAPPING = {
    "dsl.component": ("kfp", ["dsl"]),
    "dsl.pipeline": ("kfp", ["dsl"]),
    "mount_pvc": ("kfp.kubernetes", ["mount_pvc"]),
    "add_node_selector": ("kfp.kubernetes", ["add_node_selector"]),
    "client": ("kfp", ["Client"]),
    "OutputDataset": ("kfp.dsl", ["Output", "Dataset"]),
    "InputDataset": ("kfp.dsl", ["Input", "Dataset"]),
    "OutputModel": ("kfp.dsl", ["Output", "Model"]),
    "InputModel": ("kfp.dsl", ["Input", "Model"]),
}

TYPES_MAPPING = {
    "OutputDataset": "Output[Dataset]",
    "InputDataset": "Input[Dataset]",
    "OutputModel": "Output[Model]",
    "InputModel": "Input[Model]",
}

ARGPARSE_CODE = [
    "import argparse",
    "parser = argparse.ArgumentParser()",
    "parser.add_argument('-u', type=str, help='kfp URL', required=True)",
    "parser.add_argument('-c', help='Enable caching', action='store_true')",
    "parser.add_argument('-p', nargs='+', help='Placement', required=True)",
    "args = parser.parse_args()"
]

METADATA_FILENAME = "metadata.json"


import ast
import inspect
from typing import Callable, Dict

from mlopx import ComponentConverter


class Component:

    def __init__(self, image: str, func: Callable, args: Dict = None):
        self.image = image
        self.func = func
        self.name = func.__name__
        self.user_args = args
        self.arg_types = {}
        self.filename = None
        self.volumes = []

        self.get_source_file()
        self.get_tree()
        self.get_arg_types()


    def get_source_file(self) -> None:
        """
        Get the source file of the function
        """
        abs_path = inspect.getfile(self.func)
        self.filename = abs_path.split("/")[-1]

        if self.filename.split(".")[0] != self.name:
            raise ValueError("The file name must match the function name")


    def get_tree(self) -> None:
        """
        Parse the source file to an AST
        """
        with open(self.filename, "r") as f:
            self.tree = ast.parse(f.read())


    def get_arg_types(self) -> None:
        """
        Get all arguments of the function
        """
        for arg_name, arg_type in self.func.__annotations__.items():
            self.arg_types[arg_name] = arg_type.__name__


    def mount_volume(self, pvc: str, mount_path: str) -> None:
        """
        Mount a volume to a component
        """
        self.volumes.append((pvc, mount_path))


    def convert(self) -> None:
        """
        Compile the component to a kfp component
        """
        converter = ComponentConverter(self.tree)

        (
            converter.remove_type_imports()
            .add_imports(self.arg_types)
            .add_decorator(self.name, self.image)
            .update_arg_types(self.name)
            .save_component(self.filename)
        )
